<?php

namespace Drupal\camunda\Plugin\ECA\Modeler;

use Drupal;
use Drupal\Component\Plugin\PluginInspectionInterface;
use Drupal\Core\Action\ActionInterface;
use Drupal\Core\Plugin\PluginBase;
use Drupal\eca\Annotation\EcaModeler;
use Drupal\eca\Entity\Eca;
use Drupal\eca\PluginConditionInterface;
use Drupal\eca\PluginEventInterface;
use Drupal\eca\PluginModelerInterface;
use Mtownsend\XmlToArray\XmlToArray;

/**
 * Plugin implementation of the ECA Modeler.
 *
 * @EcaModeler(
 *   id = "camunda",
 * )
 */
class Camunda extends PluginBase implements PluginModelerInterface {

  /**
   * @var array
   */
  protected array $xmlModel;

  /**
   * {@inheritdoc}
   */
  public function load(string $model): PluginModelerInterface {
    $this->xmlModel = XmlToArray::convert($model);
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getId(): string {
    return $this->xmlModel['bpmn:process']['@attributes']['id'];
  }

  /**
   * {@inheritdoc}
   */
  public function getLabel(): string {
    return $this->xmlModel['bpmn:process']['@attributes']['name'] ?? 'noname';
  }

  /**
   * {@inheritdoc}
   */
  public function getStatus(): bool {
    return (bool) ($this->xmlModel['bpmn:process']['@attributes']['isExecutable'] ?? 'true');
  }

  /**
   * {@inheritdoc}
   */
  public function getVersion(): string {
    return $this->xmlModel['bpmn:process']['@attributes']['versionTag'] ?? '';
  }

  /**
   * {@inheritdoc}
   */
  public function readComponents(Eca $eca): PluginInspectionInterface {
    $flow = [];
    foreach ($this->xmlModel['bpmn:process']['bpmn:sequenceFlow'] as $sequenceFlow) {
      if (isset($sequenceFlow['bpmn:extensionElements'])) {
        $condition = $this->findAttribute($sequenceFlow, 'id');
        $eca->addCondition(
          $condition,
          $this->findProperty($sequenceFlow['bpmn:extensionElements'], 'pluginid'),
          $this->findFields($sequenceFlow['bpmn:extensionElements'])
        );
      }
      else {
        $condition = FALSE;
      }
      $flow[$this->findAttribute($sequenceFlow, 'sourceRef')][] = [
        'id' => $this->findAttribute($sequenceFlow, 'targetRef'),
        'condition' => $condition,
      ];
    }

    foreach ($this->xmlModel['bpmn:process']['bpmn:startEvent'] as $startEvent) {
      $eca->addEvent(
        $this->findAttribute($startEvent, 'id'),
        $this->findProperty($startEvent['bpmn:extensionElements'], 'pluginid'),
        $this->findAttribute($startEvent, 'name'),
        $this->findFields($startEvent['bpmn:extensionElements']),
        $flow[$this->findAttribute($startEvent, 'id')] ?? []
      );
    }

    foreach ($this->xmlModel['bpmn:process']['bpmn:task'] as $task) {
      $eca->addAction(
        $this->findAttribute($task, 'id'),
        $this->findProperty($task['bpmn:extensionElements'], 'pluginid'),
        $this->findAttribute($task, 'name'),
        $this->findFields($task['bpmn:extensionElements']),
        $flow[$this->findAttribute($task, 'id')] ?? []
      );
    }

    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function export(array $events, array $conditions, array $actions): PluginModelerInterface {
    $templates = [];
    foreach ($events as $event) {
      $templates[] = $this->prepareEventTemplate($event);
    }
    foreach ($conditions as $condition) {
      $templates[] = $this->prepareConditionTemplate($condition);
    }
    foreach ($actions as $action) {
      $templates[] = $this->prepareActionTemplate($action);
    }

    /** @noinspection PhpComposerExtensionStubsInspection */
    /** @noinspection JsonEncodingApiUsageInspection */
    file_put_contents('private://camunda.template.json', json_encode($templates));

    return $this;
  }

  /**
   * @param \Drupal\Component\Plugin\PluginInspectionInterface $plugin
   * @param string $applies_to
   * @param array $fields
   *
   * @return array
   */
  protected function properties(PluginInspectionInterface $plugin, string $applies_to, array $fields): array {
    $properties = [[
      'label' => 'Plugin ID',
      'type' => 'Hidden',
      'value' => $plugin->getPluginId(),
      'binding' => [
        'type' => 'camunda:property',
        'name' => 'pluginid',
      ],
    ]];
    foreach ($fields as $field) {
      $property = [
        'label' => $field['label'],
        'type' => $field['type'],
        'value' => $field['value'] ?? '',
        'editable' => $field['editable'] ?? TRUE,
        'binding' => [
          'type' => 'camunda:field',
          'name' => $field['name'],
        ],
      ];
      if (isset($field['extras'])) {
        /** @noinspection SlowArrayOperationsInLoopInspection */
        $property = array_merge_recursive($property, $field['extras']);
      }
      $properties[] = $property;
    }
    return [
      'name' => (string) $plugin->getPluginDefinition()['label'],
      'id' => 'org.drupal.' . $plugin->getPluginId(),
      'appliesTo' => [$applies_to],
      'properties' => $properties,
    ];
  }

  /**
   * @param \Drupal\eca\PluginEventInterface $event
   *
   * @return array
   */
  protected function prepareEventTemplate(PluginEventInterface $event): array {
    return $this->properties($event, 'bpmn:Event', $event->fields());
  }

  /**
   * @param \Drupal\eca\PluginConditionInterface $condition
   *
   * @return array
   */
  protected function prepareConditionTemplate(PluginConditionInterface $condition): array {
    return $this->properties($condition, 'bpmn:SequenceFlow', $condition->fields());
  }

  /**
   * @param \Drupal\Core\Action\ActionInterface $action
   *
   * @return array
   */
  protected function prepareActionTemplate(ActionInterface $action): array {
    return $this->properties($action, 'bpmn:Task', Drupal::service('eca.service.action')->fields($action));
  }

  /**
   * @param array $element
   * @param string $property_name
   *
   * @return string
   */
  protected function findProperty(array $element, string $property_name): string {
    $elements = isset($element['camunda:properties']['camunda:property']['@attributes']) ?
      [$element['camunda:properties']['camunda:property']] :
      $element['camunda:properties']['camunda:property'];
    /** @noinspection SuspiciousLoopInspection */
    foreach ($elements as $element) {
      if ($element['@attributes']['name'] === $property_name) {
        return $element['@attributes']['value'];
      }
    }
    return '';
  }

  /**
   * @param array $elements
   * @param string $attribute_name
   *
   * @return string
   */
  protected function findAttribute(array $element, string $attribute_name): string {
    return $element['@attributes'][$attribute_name] ?? '';
  }

  protected function findFields(array $element): array {
    $fields = [];
    if (isset($element['camunda:field'])) {
      $elements = isset($element['camunda:field']['@attributes']) ? [$element['camunda:field']] : $element['camunda:field'];
      /** @noinspection SuspiciousLoopInspection */
      foreach ($elements as $element) {
        $fields[$element['@attributes']['name']] = $element['camunda:string'];
      }
    }
    return $fields;
  }

}
